#include "WaveFileWriter.hpp"
#include "Logger.hpp"

WaveFileWriter::WaveFileWriter(WAVEFORMATEX format, size_t length, const Logger* logger)
: format_(format), length_(length), writtenLength_(0), writingLength_(0), stdout_(false), fileHandle_(NULL), logger_(logger)
{
	overlapped_.hEvent = NULL;
}

WaveFileWriter::~WaveFileWriter()
{
	close();
}

bool WaveFileWriter::open(const TCHAR* filepath)
{
	close();

	if (filepath == NULL)
	{
		// t@CpXw肳ĂȂΕWo͂ΏۂƂ
		fileHandle_ = GetStdHandle(STD_OUTPUT_HANDLE);
		if (fileHandle_ == INVALID_HANDLE_VALUE)
		{
			logger_->error(_T("Wo͂J܂BĎsĂB\n"));
			return false;
		}

		stdout_ = true;
		logger_->info(_T("o͐Wo͂ɐݒ肵܂B\n"));
	}
	else
	{
		fileHandle_ = CreateFile(filepath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (fileHandle_ == INVALID_HANDLE_VALUE)
		{
			logger_->error(_T("o̓t@CJ܂BpXmFĂB\n"));
			return false;
		}

		stdout_ = false;
		logger_->info(_T("o̓t@CJ܂B\n"));
	}

	overlapped_.Internal = 0;
	overlapped_.InternalHigh = 0;
	overlapped_.Offset = 0;
	overlapped_.OffsetHigh = 0;
	overlapped_.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (overlapped_.hEvent == NULL)
	{
		logger_->debug(_T("t@CݗpCxg̍쐬Ɏs\n"));
		logger_->error(_T("G[܂BĎsĂB\n"));
		if (!stdout_)
			CloseHandle(fileHandle_);
		fileHandle_ = NULL;
		stdout_ = false;
		return false;
	}


	// g`ݐ
	writtenLength_ = 0;


	// WAVEt@Cwb_̏o
	DWORD* header = new DWORD[11];

	// RIFFʎq
	header[0] = 0x46464952;

	// t@CTCY
	header[1] = static_cast<DWORD>(4 * 11 - 8 + length_);

	// WAVEʎq
	header[2] = 0x45564157;

	// fmt`N^O
	header[3] = 0x20746d66;

	// fmt`ÑTCY
	header[4] = 16;

	// tH[}bg^Oƃ`l
	header[5] = static_cast<DWORD>(format_.wFormatTag) | (static_cast<DWORD>(format_.nChannels) << 16);

	// TvO[g
	header[6] = static_cast<DWORD>(format_.nSamplesPerSec);

	// f[^[g
	header[7] = static_cast<DWORD>(format_.nAvgBytesPerSec);

	// ubNTCYƃrbgo
	header[8] = static_cast<DWORD>(format_.nBlockAlign) | (static_cast<DWORD>(format_.wBitsPerSample) << 16);

	// data`Ñ^Oo
	header[9] = 0x61746164;

	// g`f[^̃TCYo
	header[10] = static_cast<DWORD>(length_);

	// t@Cɏo
	BOOL result = WriteFile(fileHandle_, header, 44, NULL, &overlapped_);
	if (result == 0)
	{
		DWORD code = GetLastError();
		if (code == ERROR_IO_PENDING)
		{
			DWORD transferred;
			if (!GetOverlappedResult(fileHandle_, &overlapped_, &transferred, TRUE))
			{
				logger_->debug(_T("GetOverlappedResult֐̃G[R[h: %d\n"), GetLastError());
				logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
				close();
				return false;
			}

			if (transferred != 44)
			{
				logger_->debug(_T("f[^TCYs: v = %d, o = %d\n"), 44, transferred);
				logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
				close();
				return false;
			}
		}
		else
		{
			logger_->debug(_T("WriteFile֐̃G[R[h: %d\n"), code);
			logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
			close();
			return false;
		}
	}

	return true;
}

void WaveFileWriter::close()
{
	if (fileHandle_ != NULL)
	{
		returnWhenWritten();

		if (overlapped_.hEvent != NULL)
		{
			CloseHandle(overlapped_.hEvent);
			overlapped_.hEvent = NULL;
		}

		logger_->info(_T("f[^oI܂B\n"));
		if (!stdout_)
		{
			CloseHandle(fileHandle_);
			logger_->verbos(_T("o̓t@C܂B\n"));
		}
		fileHandle_ = NULL;
	}
}

bool WaveFileWriter::write(const void* buffer, size_t size)
{
	if (fileHandle_ == NULL)
		return false;

	// 0Byte݂̓G[Ƃ͂Ȃ
	if (size == 0)
		return true;

	// ݃TCY␳
	if (size > length_ - writtenLength_)
		size = length_ - writtenLength_;

	// oJ[\ݒ
	overlapped_.Offset = static_cast<DWORD>(writtenLength_ + 44);
#ifdef _WIN64
	// WAVEt@Ĉ4GBȏT|[gĂ炸ÃvOT|[gĂȂ̂œɈӖ͂Ȃ
	overlapped_.OffsetHigh = static_cast<DWORD>((writtenLength_ + 44) >> 32);
#endif

	// o
	BOOL result = WriteFile(fileHandle_, buffer, static_cast<DWORD>(size), NULL, &overlapped_);
	if (result == 0)
	{
		DWORD code = GetLastError();
		if (code == ERROR_IO_PENDING)
		{
			logger_->debug(_T("%d - %d ̃f[^o܂B()\n"), writtenLength_, writtenLength_ + size - 1);
			writingLength_ = size;
		}
		else
		{
			logger_->debug(_T("WriteFile֐̃G[R[h: %d\n"), code);
			logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
			return false;
		}
	}
	else
	{
		writingLength_ = 0;
		logger_->debug(_T("%d - %d ̃f[^o܂B\n"), writtenLength_, writtenLength_ + size - 1);
	}

	writtenLength_ += size;

	return true;
}

bool WaveFileWriter::returnWhenWritten()
{
	if (writingLength_ > 0)
	{
		DWORD transferred;
		if (!GetOverlappedResult(fileHandle_, &overlapped_, &transferred, TRUE))
		{
			logger_->debug(_T("GetOverlappedResult֐̃G[R[h: %d\n"), GetLastError());
			logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
			return false;
		}

		if (transferred != writingLength_)
		{
			logger_->debug(_T("f[^TCYs: v = %d, o = %d\n"), writingLength_, transferred);
			logger_->error(_T("f[^oɃG[܂BĎsĂB\n"));
			return false;
		}
	}

	return true;
}
